----------------------------------------------------------------------------------
-- Ruhr-University Bochum, Germany
-- Amir Moradi (amir.moradi@rub.de)
-- 
-- Last Update:    09:47:55 12/31/2014 
-- Module Name:    AES_EncyptionDecryption - Behavioral 
-- Project Name:   POET 2
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity AES_EncyptionDecryption is
    Port ( input  		: in  STD_LOGIC_VECTOR (127 downto 0);
           key    		: in  STD_LOGIC_VECTOR (127 downto 0);
           reset  		: in  STD_LOGIC;
           clk    		: in  STD_LOGIC;
			  encrypt		: in  STD_LOGIC;
           output 		: out STD_LOGIC_VECTOR (127 downto 0);
			  roundkeyout	: out STD_LOGIC_VECTOR (127 downto 0);
           done   		: out STD_LOGIC);
end AES_EncyptionDecryption;

architecture Behavioral of AES_EncyptionDecryption is

	COMPONENT ff_FDE IS
	PORT ( clk : IN  STD_LOGIC;
			 D   : IN  STD_LOGIC;
			 en  : IN  STD_LOGIC;
			 Q   : OUT STD_LOGIC);
	END COMPONENT;

	COMPONENT reg_SDE IS
	GENERIC ( size: POSITIVE := 128);
	PORT ( clk  : IN  STD_LOGIC;
			 D0   : IN  STD_LOGIC_VECTOR ((size-1) DOWNTO 0);
			 D1   : IN  STD_LOGIC_VECTOR ((size-1) DOWNTO 0);
			 sel  : IN  STD_LOGIC;
			 en   : IN  STD_LOGIC;
			 Q    : OUT STD_LOGIC_VECTOR ((size-1) DOWNTO 0));
	END COMPONENT;

	COMPONENT SubInvSubBytes is
	Port ( input  	: in  STD_LOGIC_VECTOR (127 downto 0);
			 encrypt	: in  STD_LOGIC;
			 output 	: out STD_LOGIC_VECTOR (127 downto 0));
	end COMPONENT;	

	COMPONENT ShiftInvShiftRows is
	Port ( input  	: in  STD_LOGIC_VECTOR (127 downto 0);
			 encrypt	: in  STD_LOGIC;
			 output 	: out STD_LOGIC_VECTOR (127 downto 0));
	end COMPONENT;	

	COMPONENT MixInvMixColumns is
	Port ( input  	: in  STD_LOGIC_VECTOR (127 downto 0);
			 encrypt	: in  STD_LOGIC;
			 output 	: out STD_LOGIC_VECTOR (127 downto 0));
	end COMPONENT;	

	COMPONENT KeyInvKeyExpansion is
	Port ( input  	: in  STD_LOGIC_VECTOR (127 downto 0);
			 encrypt	: in  STD_LOGIC;
			 Rcon   	: in  STD_LOGIC_VECTOR (7 downto 0);
			 output 	: out STD_LOGIC_VECTOR (127 downto 0));			  
	end COMPONENT;
			
	COMPONENT counter IS
	GENERIC ( size: POSITIVE);
	PORT ( clk     : IN  STD_LOGIC;
			 en      : IN  STD_LOGIC;
			 reset	: IN  STD_LOGIC;
			 q 		: OUT STD_LOGIC_VECTOR ((size-1) DOWNTO 0));
	END COMPONENT;
	
	-------------------------
	
	signal SavedEncrypt						: STD_LOGIC;
	
	signal RoundInput							: STD_LoGIC_VECTOR(127 downto 0);
	signal SubInvSubBytesInput				: STD_LoGIC_VECTOR(127 downto 0);
	signal ShiftInvShiftRowsInput			: STD_LoGIC_VECTOR(127 downto 0);
	signal ShiftInvShiftRowsOutput		: STD_LoGIC_VECTOR(127 downto 0);
	signal MixInvMixColumnsInput			: STD_LoGIC_VECTOR(127 downto 0);
	signal MixInvMixColumnsOutput			: STD_LoGIC_VECTOR(127 downto 0);
	signal RoundOutput						: STD_LoGIC_VECTOR(127 downto 0);
	
	signal SavedRoundKey						: STD_LoGIC_VECTOR(127 downto 0);
	signal KeyInvKeyExpansionOutput		: STD_LoGIC_VECTOR(127 downto 0);
	signal RoundKey1							: STD_LoGIC_VECTOR(127 downto 0);
	signal RoundKey2							: STD_LoGIC_VECTOR(127 downto 0);
	signal AddKey1								: STD_LOGIC;
	signal AddKey2								: STD_LOGIC;
	
	signal RoundEnable, LastRound			: STD_LOGIC;
	signal RoundEnableOrReset				: STD_LOGIC;
	signal RoundCounter      				: STD_LoGIC_VECTOR(3 downto 0);
	signal RconENC								: STD_LoGIC_VECTOR(7 downto 0);
	signal RconDEC								: STD_LoGIC_VECTOR(7 downto 0);
	signal Rcon									: STD_LoGIC_VECTOR(7 downto 0);
		
begin

	EncDecReg: ff_FDE
	PORT MAP( 
		clk 	=> clk,
		D   	=> encrypt,
		en  	=> reset,
		Q   	=> SavedEncrypt);

	RoundReg: reg_SDE
	GENERIC MAP (size => 128)
	PORT MAP (
		clk	=> clk,
		D0		=> RoundOutput,
		D1		=> input,
		sel	=> reset,
		en		=> RoundEnableOrReset,
		Q		=> RoundInput);
	
	SubInvSubBytesInput <= RoundInput XOR RoundKey1;
	
	SubInvSubBytesIns: SubInvSubBytes
	PORT MAP (
		input		=> SubInvSubBytesInput,
		encrypt	=> SavedEncrypt,
		output	=> ShiftInvShiftRowsInput);
	
	ShiftInvShiftRowsIns: ShiftInvShiftRows
	PORT MAP (
		input		=> ShiftInvShiftRowsInput,
		encrypt	=> SavedEncrypt,
		output	=> ShiftInvShiftRowsOutput);

	MixInvMixColumnsInput	<= ShiftInvShiftRowsOutput XOR RoundKey2;

	MixInvMixColumnsIns: MixInvMixColumns
	PORT MAP (
		input		=> MixInvMixColumnsInput,
		encrypt	=> SavedEncrypt,
		output	=> MixInvMixColumnsOutput);

	RoundOutput <= MixInvMixColumnsInput WHEN LastRound = '1' ELSE MixInvMixColumnsOutput;
	
	output 		<= SubInvSubBytesInput;
	
	-------------------------------------------------
		
	KeyReg: reg_SDE
	GENERIC MAP (size => 128)
	PORT MAP (
		clk	=> clk,
		D0		=> KeyInvKeyExpansionOutput,
		D1		=> key,
		sel	=> reset,
		en		=> RoundEnableOrReset,
		Q		=> SavedRoundKey);

	KeyInvKeyExpansionIns: KeyInvKeyExpansion
	PORT MAP (
		input		=> SavedRoundKey,
		encrypt	=> SavedEncrypt,
		Rcon		=> Rcon,
		output	=> KeyInvKeyExpansionOutput);
	
	RoundKey1	<= SavedRoundKey				 WHEN AddKey1 = '1' ELSE (others => '0');
	RoundKey2	<= KeyInvKeyExpansionOutput WHEN AddKey2 = '1' ELSE (others => '0');
		
	-------------------------------------------------

	roundkeyout	<= SavedRoundKey;

	RoundCounterIns: counter
	GENERIC MAP ( size => 4)
	PORT MAP ( 
		clk   => clk,
		en    => RoundEnable,
		reset	=> reset,
		q 		=> RoundCounter);
		
	RoundEnable 			<= '0' WHEN RoundCounter = x"A" ELSE '1';
	RoundEnableOrReset	<= RoundEnable OR reset;
	LastRound   			<= '1'   WHEN RoundCounter = x"9" ELSE '0';
	done						<= NOT RoundEnable;
	
	AddKey1					<= '1' WHEN RoundCounter = x"0" ELSE SavedEncrypt;
	AddKey2					<= NOT SavedEncrypt;
		
	Rcon						<= RconENC WHEN SavedEncrypt = '1' ELSE RconDEC;
	
	WITH RoundCounter SELECT
		RconENC <= 	x"01" WHEN x"0",
						x"02" WHEN x"1",
						x"04" WHEN x"2",
						x"08" WHEN x"3",
						x"10" WHEN x"4",
						x"20" WHEN x"5",
						x"40" WHEN x"6",
						x"80" WHEN x"7",
						x"1B" WHEN x"8",
						x"36" WHEN x"9",
						x"00" WHEN OTHERS;
		
	WITH RoundCounter SELECT
		RconDEC <= 	x"36" WHEN x"0",
						x"1B" WHEN x"1",
						x"80" WHEN x"2",
						x"40" WHEN x"3",
						x"20" WHEN x"4",
						x"10" WHEN x"5",
						x"08" WHEN x"6",
						x"04" WHEN x"7",
						x"02" WHEN x"8",
						x"01" WHEN x"9",
						x"00" WHEN OTHERS;
		
		
		
end Behavioral;

